/*
 *  FileUtils.groovy
 *  Copyright (c) 1998-2008, The University of Sheffield.
 *
 *  This code is from the GATE project (http://gate.ac.uk/) and is free
 *  software licenced under the GNU General Public License version 3. It is
 *  distributed without any warranty. For more details see COPYING.txt in the
 *  top level directory (or at http://gatewiki.sf.net/COPYING.txt).
 */

package gate.util;
import java.util.regex.Matcher
import java.util.regex.Pattern
import java.nio.charset.Charset
import java.nio.charset.IllegalCharsetNameException
import java.nio.charset.UnsupportedCharsetException
import org.apache.log4j.Logger
import org.springframework.util.StringUtils
import com.opensymphony.module.sitemesh.parser.HTMLPageParser
import com.opensymphony.module.sitemesh.HTMLPage

/**
 * Some file utilities.
 * @author Hamish Cunningham
 */
class FileUtils {
  /** 
   * Find the first occurence of a file above a location in a directory tree
   * up to and including a particular top directory.
   */
  static File findFileInTree(String fileName, File topDir, File startDir) {
    File currentDir = startDir
      log.debug("\ntopDir: ${topDir}")
    while(true) {
      log.debug("currentDir: ${currentDir}")
      File possibleFile = new File((File) currentDir, (String) fileName)
      if(possibleFile.exists())
        return possibleFile

      if(currentDir == null || topDir == null) // should never...
        return null
      else if(currentDir.getCanonicalPath() == topDir.getCanonicalPath())
        return null
      else
        currentDir = currentDir.parentFile;
    }
  } // findFileInTree

  /**
   * Set the value (content) of a DIV occuring in the BODY string of an HTML
   * file according to its ID.
   */
  static String setBodyDiv(String pageString, String divId, String text) {
    XmlParser pageParser = new XmlParser()
    def parseTree = pageParser.parseText(pageString)
    parseTree.div.each {
      if(it.attributes().id.equals(divId))
        it.value = text
    }

    def treeWriter = new StringWriter()
    new XmlNodePrinter(new PrintWriter(treeWriter)).print(parseTree)

    return treeWriter.toString()
  } // setBodyDiv(String, String, String)

  /**
   * Get the body of an HTML file. Tries to use Sitemesh (HTMLPageParser);
   * if that fails uses some imperfect string matching on body tags.
   */
  static String getBody(InputStream htmlFile) {
    String text = htmlFile.getText()
    return getBody(text)
  } // getBody(InputStream)

  /**
   * Get the body of an HTML file. Tries to use Sitemesh (HTMLPageParser);
   * if that fails uses some imperfect string matching on body tags.
   * Try to work out a correct encoding for the file.
   */
  static String getBody(File htmlFile) {
    getBodyAndMetadata(htmlFile)?.body
  }

  /**
   * Get the body and metadata from an HTML file. Tries to use Sitemesh
   * (HTMLPageParser); if that fails uses some imperfect string matching on
   * body tags.  Try to work out a correct encoding for the file.
   */
  static Map getBodyAndMetadata(File htmlFile) {
    log.debug "getBodyAndMetadata "  + htmlFile.getPath()
    String text = decodeHtmlFile(htmlFile)
    return getBodyAndMetadata(text)
  } // getBody(File)

  /**
   * Get the body of an HTML file. Tries to use Sitemesh (HTMLPageParser);
   * if that fails uses some imperfect string matching on body tags.
   */
  static String getBody(String text) {
    getBodyAndMetadata(text)?.body
  } // getBody(String)

  /**
   * Get the body and metadata from an HTML file. Tries to use Sitemesh
   * (HTMLPageParser); if that fails uses some imperfect string matching on
   * body tags.
   */
  static Map getBodyAndMetadata(String text) {
    if(text == null) return null
    if(text.length() == 0) return text;
    String body = null
    Map metadata = [:]

    // try to use sitemesh to get the page body
    HTMLPageParser parser = new HTMLPageParser()
    try {
      HTMLPage parsedPage = parser.parse(text.toCharArray())
      body = parsedPage.getBody()
      // extract the meta tags, which are added as page properties prefixed
      // with "meta." by the sitemesh parser.  We ignore "meta.http-equiv."
      // properties, which correspond to <meta http-equiv="..." content="...">
      // tags, and extract only the ones without "http-equiv.", i.e. the
      // <meta name="..." content="..."> ones
      parsedPage.propertyKeys.each { k ->
        if(k?.startsWith('meta.') && !k?.startsWith('meta.http-equiv.')) {
          metadata[k.substring(5)] = parsedPage.getProperty(k)
        }
      }
      return [body:body, metadata:metadata]
    // bug in sitemesh 2.3
    } catch(ArrayIndexOutOfBoundsException e) {
      log.debug "SiteMesh failed to parse"
    }

    // use indexOf and substring instead
    int start = text.indexOf("<body")
    if(start == -1) 
      start = text.indexOf("<BODY")
    if(start == -1) {
      log.debug "didn't find a body start tag"
      return [body:null, metadata:metadata]
    }
    int end = text.lastIndexOf("</body>")
    if(end == -1)
      end = text.lastIndexOf("</BODY>")
    if(end == -1) {
      log.debug "didn't find a body end tag"
      return null
    }
    body = text.substring(start, end)

    return [body:body, metadata:metadata]
  } // getBodyAndMetadata(String)

 /**
  * Take the html body given by a String text, and place it in the body of the
  * html File htmlFile.
  *
  * TODO: this appears not to have worked, as it used to be called by
  * PS.overwrite(Map,File,String) but didn't preserve the original HTML header
  * during edits...
  */
  static void putBody(File htmlFile, String text) {
    log.debug "putBody(${htmlFile.name}, text)"
    HTMLPageParser parser = new HTMLPageParser()
    HTMLPage parsedPage = parser.parse(decodeHtmlFile(htmlFile).toCharArray())
    String head = parsedPage?.getHead()

    // Get whatever properties we can out of the sitemesh HTMLPage
    String htmlProps = ""
    String bodProps = ""
    String titleTag = ""
    parsedPage.getProperties().each {key, value ->
      switch(key) {
        case ~/title/ :
          titleTag = "\n<title>${value}</title>\n"
          break;
        case ~/meta\.(.*)/ :
          // sitemesh puts the meta tags in the head already - do nothing
          break;
        case ~/body\.(.*)/ :
          bodProps=bodProps + "${Matcher.getLastMatcher()[0][1]}=\"${value}\" "
          break;
        default:
          htmlProps = htmlProps + "${key}=\"${value}\" "
          break;
      }
    }

    Reader reader = new StringReader(
      """<html ${htmlProps}><head>${titleTag}${head}</head>
<body ${bodProps}>
${text}
</body></html>"""
    )
    htmlFile.withWriter {it << reader}
  } // putBody(File)

  /**
   *  List all the entries in a directory, with dot files last.
   * Excludes .svn and *.yam entries.
   */
  static List dirList(File dir) {
    List entries = dir?.list().toList().sort(new DirComparator()) - ".svn"
    return entries.findAll { ! it.endsWith(".yam") }.collect {
      (new File(dir, it).getCanonicalFile().isDirectory()) ? "${it}/" : "${it}"
    }
  } // dirList(File)

  /**
   * Replace the suffix of a file name.
   * For example: on xyz.html replaces the html part; on xyz does nothing.
   */
  static String replaceFileSuffix(File file, String newSuffix) {
    return replaceFileSuffix(file.path, newSuffix)
  } // replaceFileSuffix(File, String)

  /**
   * Replace the suffix of a file name.
   * For example: on xyz.html replaces the html part; on xyz does nothing.
   */
  static String replaceFileSuffix(String path, String newSuffix) {
    if(StringUtils.getFilename(path).contains('.'))
      return path.substring(0, path.lastIndexOf('.') + 1) + newSuffix
    else
      return path
  } // replaceFileSuffix(String, String)

  /** Do some basic validity checks on a file name. */
  static boolean isValidFileName(String path) {
    if (path.trim().length() == 0) return false
    for(int i = 0; i < path.length(); i++) {
      char c = path.charAt(i)

      // is it within ASCII range (minus controls)?
      if(c < 0x20 || c > 0x7E)
        return false
    }
    return true
  } // isValidFileName(String)

  /** Logger */
  static Logger log = Logger.getLogger("gate.util.FileUtils")

  /**
   * A regex pattern used to find the encoding from an XML declaration
   * such as <?xml version="1.0" encoding="EUC-JP"?>
   */
  static Pattern XML_ENCODING_PATTERN  = ~/(?i)<\?xml[^>]*?encoding="(.+?)"/

  /**
   * A regex pattern used to find the encoding from a meta tag such as
   * <meta http-equiv="Content-type" content="text/html; charset=EUC-JP" />
   */
  static Pattern META_ENCODING_PATTERN =
       ~/(?i)<meta\s+http-equiv="Content-type"[^>]*?charset=(.+?)"/

  /**
   * Return the String contents of an html file, decoded according to the
   * encodings given in any xml declaration or meta tag. In order to overcome
   * the chicken and egg situation of reading the encoding out of a file when we
   * don't know the encoding, the method looks at the first four bytes to see if
   * it can work out the encoding. This auto encoding is then used to check for
   * an explicit encoding. The auto encoding byte check assumes an xml
   * declaration at the start of the file - it looks for the first four bytes
   * of that declaration. If there is no xml declaration, it falls back on UTF8
   * to then look for an explicit encoding. The method could be extended to
   * examine an initial  html element tag.
   * @see http://www.w3.org/TR/REC-xml/#sec-guessing-no-ext-info
   * @see http://www.w3.org/TR/REC-xml/#charencoding
   */
  static String decodeHtmlFile(File htmlFile) {

    log.debug "decodeHtmlFile "  + htmlFile.getPath()

    int fileLength = htmlFile.length()
    
    //if there are no bytes in the file then return the empty string
    if (fileLength == 0) return "";

    // Try to get the encoding. First, we look at the first four bytes,
    // and assume that it is an xml declaration
    FileInputStream fis = new FileInputStream(htmlFile)
    
    byte[] textBytes = new byte[fileLength]
    try {
      fis.read(textBytes)
    } finally {
      fis.close()
    }


    // Try to automatically determine an encoding to read in an xml declaration,
    // that we can safely use to decode the file and read the encoding attribute
    // See http://www.w3.org/TR/REC-xml/#sec-guessing-no-ext-info
    // This assumes the file starts with <?xml...
    // What if it starts with <html> or something else?  We will end up
    // using UTF-8, which is maximally failsafe. We could write a set of rules
    // for the <html> case, as with the xml rules below...

    // In the notation in comments below,
    // ## means any byte value, except two consecutive ##s cannot both be 00.
    // BOM means byte order marked. If not specified, not BOM.
    Encoding autoEncoding;
    switch(textBytes[0]) {
      case 0xFE:
        switch(textBytes[1]) {
          case 0xFF:
            // 0xFE 0xFF 0x00 0x00 	UCS-4, unusual octet order (3412), BOM
            // 0xFE 0xFF 0x## 0x## 	UTF-16, big-endian, BOM
            if(textBytes[2] == 0x00 && textBytes[3] == 0x00)
              autoEncoding = Encoding.UTF32_3412_Marked
            else
              autoEncoding == Encoding.UTF16_BE_Marked
            break;
        }
        break;
      case 0xFF:
        switch(textBytes[1]) {
          case 0xFE:
            // 0xFF 0xFE 0x00 0x00 	UCS-4, little-endian (4321 order), BOM
            // 0xFF 0xFE 0x## 0x## 	UTF-16, little-endian, BOM
            if(textBytes[2] == 0x00 && textBytes[3] == 0x00)
              autoEncoding = Encoding.UTF32_LE_Marked
            else
              autoEncoding == Encoding.UTF16_LE_Marked
            break;
        }
        break;
      case 0xEF:
        //0xEF 0xBB 0xBF 0x##	  UTF-8, BOM
        if(textBytes[1] == 0xBB && textBytes[2] == 0xBF)
          autoEncoding == Encoding.UTF8;
        break;
      case 0x00:
        switch(textBytes[1]) {
          case 0x00:
            // 0x00 0x00 0xFE 0xFF  	UCS-4, big-endian machine (1234 order) BOM
            if(textBytes[2] == 0xFE && textBytes[3] == 0xFF)
              autoEncoding = Encoding.UTF32_BE_Marked
            // 0x00 0x00 0xFF 0xFE 	UCS-4, unusual octet order (2143), BOM
            else if(textBytes[2] == 0xFF && textBytes[3] == 0xFE)
              autoEncoding = Encoding.UTF32_2143_Marked
            // 0x00 0x00 0x00 0x3C  	UCS-4 etc, big-endian (1234 order)
            else if(textBytes[2] == 0x00 && textBytes[3] == 0x3C)
              autoEncoding = Encoding.UTF32_BE_Unmarked
            // 0x00 0x00 0x3C 0x00  	UCS-4 etc, unusual order (2143)
            else if(textBytes[2] == 0x3C && textBytes[3] == 0x00)
              autoEncoding = Encoding.UTF32_2143_Unmarked
            break;
          case 0x3C:
            // 0x00 0x3C 0x00 0x00  	UCS-4 etc, unusual order (3412)
            if(textBytes[2] == 0x00 && textBytes[3] == 0x00)
              autoEncoding = Encoding.UTF32_3412_Unmarked
            // 0x00 0x3C 0x00 0x3F 	UTF-16BE or big-endian ISO-10646-UCS-2 etc
            else if(textBytes[2] == 0x00 && textBytes[3] == 0x3F)
              autoEncoding = Encoding.UTF16_BE_Unmarked
            break;
        }
        break;
      case 0x3C:
        switch(textBytes[1]) {
          case 0x00:
            // 0x3C 0x00 0x00 0x00  	UCS-4 etc, little-endian (4321)
            if(textBytes[2] == 0x00 && textBytes[3] == 0x00)
              autoEncoding = Encoding.UTF32_LE_Unmarked
            // 0x3C 0x00 0x3F 0x00 	UTF-16LE, little-endian ISO-10646-UCS-2 etc
            else if(textBytes[2] == 0x3F && textBytes[3] == 0x00)
              autoEncoding = Encoding.UTF16_LE_Unmarked
            break;
          case 0x3F:
            // 0x3C 0x3F 0x78 0x6D 	UTF-8, ISO 646, ASCII, etc
            if(textBytes[2] == 0x78 && textBytes[3] == 0x6D)
              autoEncoding = Encoding.UTF8
            break;
        }
        break;
      case 0x4C:
        // 0x4C 0x6F 0xA7 0x94 	EBCDIC (in some flavor)
        if(textBytes[1] == 0x6F && textBytes[2] == 0xA7 && textBytes[3] == 0x94)
          autoEncoding = Encoding.EBCDIC
        break;
    }

    // If we don't have an automatic encoding now, it is either mislabelled,
    // not xml, wrapped, or UTF-8. Try UTF-8 - the safest bet. We use UTF-8
    // because (a) it is maximally failsafe (b) gives us uniform treatment
    // across all platforms, and (c) it is the only allowed encoding in XML
    // that has no encoding in its declaration
    // (see http://www.w3.org/TR/REC-xml/#charencoding)
    if(autoEncoding == null) {
// TODO cope better with HTML files and switch back to warn
      log.debug("auto encoding not found from first 4 bytes, trying "
               + "${Encoding.UTF8.encoding()}")
      autoEncoding = Encoding.UTF8
    } else if(!autoEncoding.supported()) {
      // If we have an encoding but it is not supported, try UTF-8
// TODO cope better with HTML files and switch back to warn
      log.debug("unsupported auto encoding in first 4 bytes: "
              + "${autoEncoding.encoding()}, trying ${Encoding.UTF8.encoding()}"
              +", decoding might fail")
      autoEncoding = Encoding.UTF8
    } else if(autoEncoding.partial()) {
// TODO cope better with HTML files and switch back to warn
      log.debug("using auto encoding with partial support: "
              + "${autoEncoding.encoding()}, decoding might fail")            
    } else {
        log.debug "auto encoding from first 4 bytes: ${autoEncoding.encoding()}"
    }

    // Now we have a "safe" encoding for reading in. Look for the encoding
    // attribute in the xml declaration, and if that fails, look for an html
    // meta element. The XML element takes precedence.
    // If we have still failed, use 8859-1, which is considered the default
    // by html / http standards.

    // We could perhaps avoid decoding the whole array here...
    // We could perhaps only grep up to a <body> tag

    String autoString = new String(textBytes, autoEncoding.encoding())
    String encoding;
    Matcher xmlMatcher = XML_ENCODING_PATTERN.matcher(autoString)
    if(xmlMatcher.find()) {
      encoding = xmlMatcher[0][1]
      log.debug "found encoding in xml declaration: ${encoding}"
    } else {
      Matcher metaMatcher = META_ENCODING_PATTERN.matcher(autoString)
      if(metaMatcher.find()){
        encoding = metaMatcher[0][1]
        log.debug "found encoding in meta tag: ${encoding}"
      } else {
        encoding = Encoding.HTML_default.encoding()
        log.debug "no encoding found, using html default encoding: ${encoding}"
      }
    }

    // Is this encoding supported? Get its canonical name, for comparison
    // to auto encoding? If not supported, use the html default.
    try{
      encoding = Charset.forName(encoding).name()
      log.debug "using canonical encoding: ${encoding}"
    } catch(IllegalCharsetNameException icne) {
      log.debug("illegal encoding name: ${encoding}, "
                + "using ${Encoding.HTML_default.encoding()} instead")
      encoding = Encoding.HTML_default.encoding()
    } catch(UnsupportedCharsetException uce) {
      encoding = Encoding.HTML_default.encoding()
      log.debug("unsupported encoding: ${encoding}, "
                + "using ${Encoding.HTML_default.encoding()} instead")
    }

    // Now get the chars, using this encoding, and get the body from them.
    // Reread the file if neccesary
    if(encoding.equals(autoEncoding.encoding()))
      return autoString
    else
      return new String(textBytes, encoding)

  } // decodeHtmlFile

} // FileUtils

/**
 * A comparator for directory listings.
 */
class DirComparator implements Comparator {
  int compare(a, b) {
    if(a.startsWith(".") && !b.startsWith("."))
      return 1
    else if(b.startsWith(".") && !a.startsWith("."))
      return -1
    else
      return a.toLowerCase().compareTo(b.toLowerCase())
  } // compare
} // DirComparator

/**
 * An enumeration of automatic encodings that can be determined from XML files.
 * Note that this si not the same as the encodings and charsets found in the
 * java api, hence the need for a separate enumeration.
 * Each Encoding has an encoding (the canonical name from the
 * java.nio API, and a boolean supported flag, which is true if the encoding is
 * supported. Each also has a partial flag, which is true if support is partial.
 */
public enum Encoding {

  HTML_default        ("ISO-8859-1",             true),
  UTF8                ("UTF-8",                  true),
  // Note that for EBCDIC IBM1047 *might* work sometimes, but there are
  // other EBCDIC flavours that will break, and other things to consider.
  // There are some libraries available for EBCDIC to ASCII conversion.
  // if there is ever the need to support it fully...
  EBCDIC              ("IBM1047",                true, true),
  UTF16_BE_Marked     ("UTF-16",                 true),
  UTF16_LE_Marked     ("UTF-16",                 true),
  UTF16_BE_Unmarked   ("UTF-16BE",               true),
  UTF16_LE_Unmarked   ("UTF-16LE",               true),
  // Java 1.5 does not support UTF32. Java 1.6 does.
  UTF32_BE_Marked     ("x-UTF-32BE-BOM",         false),
  UTF32_BE_Unmarked   ("UTF-32BE",               false),
  UTF32_LE_Marked     ("x-UTF-32LE-BOM",         false),
  UTF32_LE_Unmarked   ("UTF-32LE",               false),
  UTF32_2143_Marked   ("UTF-32",                 false),
  UTF32_2143_Unmarked ("UTF-32",                 false),
  UTF32_3412_Marked   ("UTF-32",                 false),
  UTF32_3412_Unmarked ("UTF-32",                 false);

  /** The encoding: java.nio API canonical name*/
  private final String encoding;

  /** Is this Encoding supoprted?*/
  private final boolean supported;

  /** Is support for this encoding partial?*/
  private final boolean partial;

  /**
   * Create an Encoding with the given encoding, supported and warn flags.
   */
  Encoding(String encoding, boolean supported, boolean partial) {
    this.encoding = encoding
    this.supported = supported
    this.partial = partial
  }

  /**
   * Create an Encoding with the given encoding and supported flag. Set the
   * partial flag to true if supported is false, otherwise to false
   */
  Encoding(String encoding, boolean supported) {
    this.encoding = encoding
    this.supported = supported
    this.partial = !supported
  }

  /** Return the encoding String of this encoding,
   * the java.nio API canonical name
   */
  public String encoding() { return encoding }

  /** Return the supported flag of this encoding */
  public boolean supported() { return supported }

  /** Return the partial flag of this encoding */
  public boolean partial() { return partial }

} // Encoding
